from __future__ import print_function
import torch
import torch.nn as nn
import torch.nn.functional as F
import torch.optim as optim

import copy
import random
import numpy as np
from numpy import linalg as LA
import math
import yaml

def add_sparse_args(parser):
    parser.add_argument('--sparse', action='store_true', help='Enable sparse mode. Default: True.')
    parser.add_argument('--fix', action='store_true', help='Fix sparse connectivity during training. Default: True.')
    parser.add_argument('--sparse_init', type=str, default='ERK', help='sparse initialization')
    parser.add_argument('--growth', type=str, default='ucb', help='Growth mode. Choose from: momentum, random, random_unfired, gradient and ucb.')
    parser.add_argument('--death', type=str, default='magnitude', help='Death mode / pruning mode. Choose from: magnitude, SET, threshold.')
    parser.add_argument('--redistribution', type=str, default='none', help='Redistribution mode. Choose from: momentum, magnitude, nonzeros, or none.')
    parser.add_argument('--death-rate', type=float, default=0.4, help='The pruning rate / death rate.')
    parser.add_argument('--density', type=float, default=0.05, help='The density of the overall sparse network.')
    parser.add_argument('--update_frequency', type=int, default=1, metavar='N', help='how many iterations to train between parameter exploration')
    parser.add_argument('--decay-schedule', type=str, default='cosine', help='The decay schedule for the pruning rate. Default: cosine. Choose from: cosine, linear.')
    parser.add_argument('--theta', type=float, default=1e-5, help='upper confidence bound coefficient')
    parser.add_argument('--theta_decay_freq', type=float, default=400, help='theta decay frequency')
    parser.add_argument('--epsilon', type=float, default=1.0, help='')
    parser.add_argument('--factor', type=float, default=1, help='')

class CosineDecay(object):
    def __init__(self, death_rate, T_max, eta_min=0.005, last_epoch=-1):
        self.sgd = optim.SGD(torch.nn.ParameterList([torch.nn.Parameter(torch.zeros(1))]), lr=death_rate)
        self.cosine_stepper = torch.optim.lr_scheduler.CosineAnnealingLR(self.sgd, T_max, eta_min, last_epoch)

    def step(self):
        self.cosine_stepper.step()

    def get_dr(self):
        return self.sgd.param_groups[0]['lr']

# theta cosine decay
class CosineDecayTheta(object):
    def __init__(self, theta, T_max, eta_min=1e-30, last_epoch=-1):
        print("T_max: ", T_max)
        print("eta_min: ", eta_min)
        self.sgd = optim.SGD(torch.nn.ParameterList([torch.nn.Parameter(torch.zeros(1))]), lr=theta)
        self.cosine_stepper = torch.optim.lr_scheduler.CosineAnnealingLR(self.sgd, T_max, eta_min, last_epoch)

    def step(self):
        self.cosine_stepper.step()

    def get_dr(self):
        return self.sgd.param_groups[0]['lr']


class LinearDecay(object):
    def __init__(self, death_rate, factor=0.1, frequency=4000):
        self.factor = factor
        self.steps = 0
        self.frequency = frequency

    def step(self):
        self.steps += 1

    def get_dr(self):
        if self.steps > 0 and self.steps % self.frequency == 0:
            return self.death_rate*self.factor
        else:
            return self.death_rate


class LinearDecayTheta(object):
    def __init__(self, theta, factor=1, theta_decay_freq=4000):
        self.theta = theta
        self.factor = factor
        self.steps = 0
        self.theta_decay_freq = theta_decay_freq

    def step(self):
        self.steps += 1

    def get_dr(self, theta, factor):
        if self.steps > 0 and self.steps % self.theta_decay_freq == 0:
            return theta*self.factor
        else:
            return theta


class Masking(object):
    def __init__(self, optimizer, death_rate=0.3, growth_death_ratio=1.0, death_rate_decay=None, theta_decay=None, death_mode='magnitude', growth_mode='momentum', redistribution_mode='momentum', threshold=0.001, theta=1e-5, epsilon=1.0, factor=1, args=None):
        growth_modes = ['random', 'momentum', 'momentum_neuron', 'gradient', 'row_balanced_gradient', 'row_wise_gradient']

        self.args = args
        self.device = args.device
        self.growth_mode = growth_mode
        self.death_mode = death_mode
        self.growth_death_ratio = growth_death_ratio
        self.redistribution_mode = redistribution_mode
        self.death_rate_decay = death_rate_decay

        self.masks = {}
        self.modules = []
        self.names = []
        self.optimizer = optimizer

        # stats
        self.name2zeros = {}
        self.num_remove = {}
        self.name2nonzeros = {}
        self.death_rate = death_rate
        self.theta = theta
        self.theta_decay = theta_decay
        self.factor = factor
        self.epsilon = epsilon
        self.baseline_nonzero = None
        self.steps = 0
        self.num_pruned_rows = {}
        self.num_unpruned_rows = {}
        self.num_rows_remove = {}
        self.fired_times = {}
        self.grad_total = {}
        self.grad_mov_avg ={}


        if self.args.fix: self.prune_every_k_steps = None
        else: self.prune_every_k_steps = self.args.update_frequency

    def init(self, mode='ERK', density=0.05, erk_power_scale=1.0):
        self.density = density
        if mode == 'GMP':
            self.baseline_nonzero = 0
            for module in self.modules:
                for name, weight in module.named_parameters():
                    if name not in self.masks: continue
                    self.masks[name] = torch.ones_like(weight, dtype=torch.float32, requires_grad=False).to(self.device)
                    self.baseline_nonzero += (self.masks[name] != 0).sum().int().item()

        elif mode == 'lottery_ticket':
            self.baseline_nonzero = 0
            weight_abs = []
            for module in self.modules:
                for name, weight in module.named_parameters():
                    if name not in self.masks: continue
                    weight_abs.append(torch.abs(weight))

            # Gather all scores in a single vector and normalise
            all_scores = torch.cat([torch.flatten(x) for x in weight_abs])
            num_params_to_keep = int(len(all_scores) * self.density)

            threshold, _ = torch.topk(all_scores, num_params_to_keep, sorted=True)
            acceptable_score = threshold[-1]

            for module in self.modules:
                for name, weight in module.named_parameters():
                    if name not in self.masks: continue
                    self.masks[name] = ((torch.abs(weight)) >= acceptable_score).float()
                    self.baseline_nonzero += (self.masks[name]!=0).sum().int().item()

        elif mode == 'uniform':
            self.baseline_nonzero = 0
            for module in self.modules:
                for name, weight in module.named_parameters():
                    if name not in self.masks: continue
                    self.masks[name][:] = (torch.rand(weight.shape) < density).float().data.to(self.device) 
                    self.baseline_nonzero += weight.numel()*density

        elif mode == 'row_balanced_uniform':
            self.baseline_nonzero = 0
            for module in self.modules:
                for name, weight in module.named_parameters():
                    if name not in self.masks: continue
                    weight = weight.cpu().detach().numpy()
                    self.masks[name][:] = torch.rand(weight.shape)

                    for i in range(weight.shape[0]):
                        y, idx = torch.sort(torch.abs(self.masks[name][i,:]).flatten())  # sort by ascending
                        num_zeros_per_row = int(weight.shape[-1]*(1-density))
                        self.masks[name][i,:].view(-1)[idx[:num_zeros_per_row]] = 0.0
                        self.masks[name][i,:] = (self.masks[name][i,:] != 0.0)
                        
                    self.masks[name] = self.masks[name].float().data.to(self.device)
                    weight = torch.from_numpy(weight).float()
                   
                    self.baseline_nonzero += (self.masks[name]!=0).sum().int().item()
                  
            for module in self.modules:
                for name, weight in module.named_parameters():
                    if name not in self.masks: continue
                    mask = self.masks[name]

                    self.name2nonzeros[name] = mask.sum().item()
                    self.name2zeros[name] = mask.numel() - self.name2nonzeros[name]

        elif mode == 'row_wise_uniform':
            self.baseline_nonzero = 0
            for module in self.modules:
                for name, weight in module.named_parameters():
                    if name not in self.masks: continue
                    weight = weight.cpu().detach().numpy()
                    shape = weight.shape
                    weight2d = weight.reshape(shape[0], -1)
                    self.masks[name][:] = torch.rand(weight.shape)
                    num_zeros_rows = int(weight2d.shape[0]*(1-density))

               
                    row_l2_norm = LA.norm(weight2d, 2, axis=1)
                    row_l2_norm_sorted = np.argsort(row_l2_norm)
                    self.num_pruned_rows[name] = int(weight.shape[0]*(1-density))
                    self.num_unpruned_rows[name] = weight.shape[0] - self.num_pruned_rows[name]

                    for i in range(self.num_pruned_rows[name]):
                        self.masks[name][row_l2_norm_sorted[i],:] = 0.0
                        
                    self.masks[name] = self.masks[name].float().data.to(self.device)
                    weight = torch.from_numpy(weight).float()
                  
                    self.baseline_nonzero += (self.masks[name]!=0).sum().int().item()
      
            for module in self.modules:
                for name, weight in module.named_parameters():
                    if name not in self.masks: continue
                    mask = self.masks[name]

                    self.name2nonzeros[name] = mask.sum().item()
                    self.name2zeros[name] = mask.numel() - self.name2nonzeros[name]

                

        elif mode == 'ERK':
            # print('initialize by ERK')
            total_params = 0
            for name, weight in self.masks.items():
                total_params += weight.numel()
            is_epsilon_valid = False
            # # The following loop will terminate worst case when all masks are in the
            # custom_sparsity_map. This should probably never happen though, since once
            # we have a single variable or more with the same constant, we have a valid
            # epsilon. Note that for each iteration we add at least one variable to the
            # custom_sparsity_map and therefore this while loop should terminate.
            dense_layers = set()
            while not is_epsilon_valid:
                # We will start with all layers and try to find right epsilon. However if
                # any probablity exceeds 1, we will make that layer dense and repeat the
                # process (finding epsilon) with the non-dense layers.
                # We want the total number of connections to be the same. Let say we have
                # for layers with N_1, ..., N_4 parameters each. Let say after some
                # iterations probability of some dense layers (3, 4) exceeded 1 and
                # therefore we added them to the dense_layers set. Those layers will not
                # scale with erdos_renyi, however we need to count them so that target
                # paratemeter count is achieved. See below.
                # eps * (p_1 * N_1 + p_2 * N_2) + (N_3 + N_4) =
                #    (1 - default_sparsity) * (N_1 + N_2 + N_3 + N_4)
                # eps * (p_1 * N_1 + p_2 * N_2) =
                #    (1 - default_sparsity) * (N_1 + N_2) - default_sparsity * (N_3 + N_4)
                # eps = rhs / (\sum_i p_i * N_i) = rhs / divisor.

                divisor = 0
                rhs = 0
                raw_probabilities = {}
                for name, mask in self.masks.items():
                    n_param = np.prod(mask.shape)
                    n_zeros = n_param * (1 - self.density)
                    n_ones = n_param * self.density

                    if name in dense_layers:
                        rhs -= n_zeros
                    else:
                        rhs += n_ones
                        raw_probabilities[name] = (
                                                          np.sum(mask.shape) / np.prod(mask.shape)
                                                  ) ** erk_power_scale
                        divisor += raw_probabilities[name] * n_param
                epsilon = rhs / divisor
                max_prob = np.max(list(raw_probabilities.values()))
                max_prob_one = max_prob * epsilon
                if max_prob_one > 1:
                    is_epsilon_valid = False
                    for mask_name, mask_raw_prob in raw_probabilities.items():
                        if mask_raw_prob == max_prob:
                            # print(f"Sparsity of var:{mask_name} had to be set to 0.")
                            dense_layers.add(mask_name)
                else:
                    is_epsilon_valid = True

            density_dict = {}
            total_nonzero = 0.0
            # With the valid epsilon, we can set sparsities of the remaning layers.
            for name, mask in self.masks.items():
                n_param = np.prod(mask.shape)
                if name in dense_layers:
                    density_dict[name] = 1.0
                else:
                    probability_one = epsilon * raw_probabilities[name]
                    density_dict[name] = probability_one
            
                self.masks[name][:] = (torch.rand(mask.shape) < density_dict[name]).float().data.to(self.device)

                total_nonzero += density_dict[name] * mask.numel()
       
        
        self.apply_mask()
        self.fired_masks = copy.deepcopy(self.masks) # used for ITOP
        self.fired_times = copy.deepcopy(self.masks)

        total_size = 0
        for name, weight in self.masks.items():
            total_size  += weight.numel()
    
        sparse_size = 0
        for name, weight in self.masks.items():
            sparse_size += (weight != 0).sum().int().item()

      
    def step(self):
        self.optimizer.step()
        self.apply_mask()
        self.death_rate_decay.step()
        self.death_rate = self.death_rate_decay.get_dr()
        self.theta_decay.step()
        self.theta = self.theta_decay.get_dr(self.theta, self.factor)
        self.steps += 1


        if self.prune_every_k_steps is not None:
            if self.steps % self.prune_every_k_steps == 0:
                self.truncate_weights()
                _, _ = self.fired_masks_update()
                # self.print_nonzero_counts()


    def add_module(self, module, density, sparse_init='ER', weight_names=None):
       
        self.modules.append(module)

        for name, tensor in module.named_parameters():
            if name in weight_names:
                self.names.append(name)
                self.masks[name] = torch.zeros_like(tensor, dtype=torch.float32, requires_grad=False).to(self.device)

  
        self.init(mode=sparse_init, density=density)


    def remove_weight(self, name):
        if name in self.masks:
         
            self.masks.pop(name)
        elif name + '.weight' in self.masks:
        
            self.masks.pop(name + '.weight')
        else:
            print('ERROR', name)

    def remove_weight_partial_name(self, partial_name):
        removed = set()
        for name in list(self.masks.keys()):
            if partial_name in name:

                removed.add(name)
                self.masks.pop(name)

     
        i = 0
        while i < len(self.names):
            name = self.names[i]
            if name in removed:
                self.names.pop(i)
            else:
                i += 1

    def remove_type(self, nn_type):
        for module in self.modules:
            for name, module in module.named_modules():
                if isinstance(module, nn_type):
                    self.remove_weight(name)

    def apply_mask(self):
        for module in self.modules:
            for name, tensor in module.named_parameters():
                if name in self.masks:
                    tensor.data = tensor.data*self.masks[name]
                    if 'momentum_buffer' in self.optimizer.state[tensor]:
                        self.optimizer.state[tensor]['momentum_buffer'] = self.optimizer.state[tensor]['momentum_buffer']*self.masks[name]

    def truncate_weights_GMP(self, epoch):
        '''
        Implementation  of GMP To prune, or not to prune: exploring the efficacy of pruning for model compression https://arxiv.org/abs/1710.01878
        :param epoch: current training epoch
        :return:
        '''
        prune_rate = 1 - self.density
        curr_prune_epoch = epoch
        total_prune_epochs = self.args.multiplier * self.args.final_prune_epoch - self.args.multiplier * self.args.init_prune_epoch + 1
        if epoch >= self.args.multiplier * self.args.init_prune_epoch and epoch <= self.args.multiplier * self.args.final_prune_epoch:
            prune_decay = (1 - ((curr_prune_epoch - self.args.multiplier * self.args.init_prune_epoch) / total_prune_epochs)) ** 3
            curr_prune_rate = prune_rate - (prune_rate * prune_decay)

            for module in self.modules:
                for name, weight in module.named_parameters():
                    if name not in self.masks: continue

                    x, idx = torch.sort(torch.abs(weight.data.view(-1)))
                    p = int(curr_prune_rate * weight.numel())
                    self.masks[name].data.view(-1)[idx[:p]] = 0.0
            self.apply_mask()
        total_size = 0
        for name, weight in self.masks.items():
            total_size += weight.numel()
      
        sparse_size = 0
        for name, weight in self.masks.items():
            sparse_size += (weight != 0).sum().int().item()

    
    def truncate_weights(self):

        fired_times = {}
      
        for module in self.modules:
            for name, weight in module.named_parameters():
                if name not in self.masks: continue
                mask = self.masks[name]

                self.name2nonzeros[name] = mask.sum().item()
                self.name2zeros[name] = mask.numel() - self.name2nonzeros[name]

                # death
                if self.death_mode == 'magnitude':
                    new_mask = self.magnitude_death(mask, weight, name)
                elif self.death_mode == 'row_balanced':
                    new_mask = self.row_balanced_death(mask, weight, name)
                elif self.death_mode == 'row_wise':
                    new_mask = self.row_wise_death(mask, weight, name)
                elif self.death_mode == 'SET':
                    new_mask = self.magnitude_and_negativity_death(mask, weight, name)
                elif self.death_mode == 'Taylor_FO':
                    new_mask = self.taylor_FO(mask, weight, name)
                elif self.death_mode == 'threshold':
                    new_mask = self.threshold_death(mask, weight, name)

                self.num_remove[name] = int(self.name2nonzeros[name] - new_mask.sum().item())
              
                self.masks[name][:] = new_mask
                self.name2nonzeros[name] = mask.sum().item()
                self.name2zeros[name] = mask.numel() - self.name2nonzeros[name]
             
        # print("*********************************After drop*********************************")
        for module in self.modules:
            for name, weight in module.named_parameters():
                if name not in self.masks: continue
                mask = self.masks[name]

                self.name2nonzeros[name] = mask.sum().item()
                self.name2zeros[name] = mask.numel() - self.name2nonzeros[name]

            
        for module in self.modules:
            for name, weight in module.named_parameters():
                if name not in self.masks: continue
                new_mask = self.masks[name].data.byte()

                # growth
                if self.growth_mode == 'random':
                    new_mask = self.random_growth(name, new_mask, weight)

                if self.growth_mode == 'random_unfired':
                    new_mask = self.random_unfired_growth(name, new_mask, weight)

                elif self.growth_mode == 'momentum':
                    new_mask = self.momentum_growth(name, new_mask, weight)

                elif self.growth_mode == 'gradient':
                    new_mask = self.gradient_growth(name, new_mask, weight)

                elif self.growth_mode == 'row_balanced_gradient':
                    new_mask = self.row_balanced_gradient_growth(name, new_mask, weight)

                elif self.growth_mode == 'row_wise_gradient':
                    new_mask = self.row_wise_gradient_growth(name, new_mask, weight)

                elif self.growth_mode == 'ucb':
                    new_mask = self.ucb_growth(name, new_mask, weight)
                    self.fired_times[name] = self.fired_times[name] + new_mask.float()
                
                elif self.growth_mode == 'ucb_row_wise':
                    new_mask = self.ucb_growth_row(name, new_mask, weight)
                    self.fired_times[name] = self.fired_times[name] + new_mask.float()
     
                # exchanging masks
                self.masks.pop(name)
            
                self.masks[name] = new_mask.float()


        self.apply_mask()


    '''
                    DEATH
    '''

    def threshold_death(self, mask, weight, name):
        return (torch.abs(weight.data) > self.threshold)

    def taylor_FO(self, mask, weight, name):

        num_remove = math.ceil(self.death_rate * self.name2nonzeros[name])
        num_zeros = self.name2zeros[name]
        k = math.ceil(num_zeros + num_remove)

        x, idx = torch.sort((weight.data * weight.grad).pow(2).flatten())
        mask.data.view(-1)[idx[:k]] = 0.0

        return mask

    def magnitude_death(self, mask, weight, name):
        num_remove = math.ceil(self.death_rate*self.name2nonzeros[name])
        if num_remove == 0.0: return weight.data != 0.0
        num_zeros = self.name2zeros[name]
        
        x, idx = torch.sort(torch.abs(weight.data.view(-1)))
        n = idx.shape[0]
        k = math.ceil(num_zeros + num_remove)
        threshold = x[k-1].item()

        return (torch.abs(weight.data) > threshold)

    def row_balanced_death(self, mask, weight, name):

        num_remove_per_row = math.ceil(self.death_rate*self.name2nonzeros[name]/weight.shape[0])

        if num_remove_per_row == 0.0: return weight.data != 0.0
        num_zeros = self.name2zeros[name]

        shape = weight.shape
        weight2d = weight.reshape(shape[0], -1)
        shape2d = weight2d.shape
        new_mask = torch.zeros([shape2d[0], shape2d[1]])
        k = math.ceil((num_zeros /shape2d[0])) + num_remove_per_row
     
        for i in range(shape2d[0]):
            y, idx = torch.sort(torch.abs(weight2d.data[i,:]))
            threshold = y[k-1].item()
            new_mask[i, :] = (torch.abs(weight2d.data[i,:]) > threshold)
        return new_mask

    def row_wise_death(self, mask, weight, name):
        self.num_rows_remove[name] = math.ceil(self.death_rate*self.num_unpruned_rows[name])
        num_rows_remained = self.num_unpruned_rows[name] - self.num_rows_remove[name]
        


        weight = weight.cpu().detach().numpy()
        shape = weight.shape
        weight2d = weight.reshape(shape[0], -1)
        shape2d = weight2d.shape
        new_mask = torch.zeros([shape2d[0], shape2d[1]])
   
        row_l2_norm = LA.norm(weight2d, 2, axis=1)
        row_l2_norm_sorted = np.argsort(row_l2_norm)[::-1][:len(row_l2_norm)]

        for i in range(num_rows_remained):
            
            new_mask[row_l2_norm_sorted[i], :] = 1.0
        return new_mask

    def pattern_magnitude_death(self, mask, weight, name):
        # TODO
        return

    def magnitude_and_negativity_death(self, mask, weight, name):
        num_remove = math.ceil(self.death_rate*self.name2nonzeros[name])
        num_zeros = self.name2zeros[name]

        x, idx = torch.sort(weight[weight > 0.0].data.view(-1))
        k = math.ceil(num_remove/2.0)
        if k >= x.shape[0]:
            k = x.shape[0]

        threshold_magnitude = x[k-1].item()

        x, idx = torch.sort(weight[weight < 0.0].view(-1))
        k = math.ceil(num_remove/2.0)
        if k >= x.shape[0]:
            k = x.shape[0]
        threshold_negativity = x[k-1].item()

        pos_mask = (weight.data > threshold_magnitude) & (weight.data > 0.0)
        neg_mask = (weight.data < threshold_negativity) & (weight.data < 0.0)


        new_mask = pos_mask | neg_mask
        return new_mask

    '''
                    GROWTH
    '''

    def random_unfired_growth(self, name, new_mask, weight):
        total_regrowth = self.num_remove[name]
        n = (new_mask == 0).sum().item()
        if n == 0: return new_mask
        num_nonfired_weights = (self.fired_masks[name]==0).sum().item()

        if total_regrowth <= num_nonfired_weights:
            idx = (self.fired_masks[name].flatten() == 0).nonzero()
            indices = torch.randperm(len(idx))[:total_regrowth]

        
            new_mask.data.view(-1)[idx[indices]] = 1.0
        else:
            new_mask[self.fired_masks[name]==0] = 1.0
            n = (new_mask == 0).sum().item()
            expeced_growth_probability = ((total_regrowth-num_nonfired_weights) / n)
            new_weights = torch.rand(new_mask.shape).to(self.device) < expeced_growth_probability
            new_mask = new_mask.byte() | new_weights
        return new_mask

    def random_growth(self, name, new_mask, weight):
        total_regrowth = self.num_remove[name]
        n = (new_mask==0).sum().item()
        if n == 0: return new_mask
        expeced_growth_probability = (total_regrowth/n)
        new_weights = torch.rand(new_mask.shape).to(self.device) < expeced_growth_probability
        new_mask_ = new_mask.byte() | new_weights
        if (new_mask_!=0).sum().item() == 0:
            new_mask_ = new_mask
        return new_mask_

    def momentum_growth(self, name, new_mask, weight):
        total_regrowth = self.num_remove[name]
        grad = self.get_momentum_for_weight(weight)
        grad = grad*(new_mask==0).float()
        y, idx = torch.sort(torch.abs(grad).flatten(), descending=True)
        new_mask.data.view(-1)[idx[:total_regrowth]] = 1.0
        return new_mask

    
    def ucb_growth(self, name, new_mask, weight):
        total_regrowth = self.num_remove[name]
        grad = self.get_gradient_for_weights(weight)
    
        t = self.steps
        N_weight = self.fired_times[name]
        ucb = (grad + self.theta * math.log(t) / (N_weight + self.epsilon))*(new_mask==0).float()

        y, idx = torch.sort(torch.abs(ucb).flatten(), descending=True)
        new_mask.data.view(-1)[idx[:total_regrowth]] = 1.0
        return new_mask


    def ucb_growth_row(self, name, new_mask, weight):
        self.num_rows_regrowth = self.num_rows_remove[name]
        grad = self.get_gradient_for_weights(weight)
       
        t = self.steps

        row_l2_norm_grad = LA.norm(grad.cpu().detach().numpy(), 2, axis=1)
        row_l2_norm_grad = torch.from_numpy(row_l2_norm_grad).to(self.device)
       
        N_weight = self.fired_times[name][:,0] 
        new_mask_one_dim = new_mask[:, 0]

        ucb_one_dim = (row_l2_norm_grad + self.theta * math.log(t) / (N_weight + self.epsilon))*(new_mask_one_dim==0).float()
        ucb_one_dim = ucb_one_dim.cpu().detach().numpy()
        ucb_sorted = np.argsort(ucb_one_dim)[::-1][:len(ucb_one_dim)]
   
        for i in range(self.num_rows_regrowth):
            new_mask[ucb_sorted[i], :] = 1.0

        return new_mask

    

    def gradient_growth(self, name, new_mask, weight):
        total_regrowth = self.num_remove[name]
        grad = self.get_gradient_for_weights(weight)
      
        grad = grad*(new_mask==0).float()
        

        y, idx = torch.sort(torch.abs(grad).flatten(), descending=True)
   
        new_mask.data.view(-1)[idx[:total_regrowth]] = 1.0
        return new_mask


    def row_wise_gradient_growth(self, name, new_mask, weight):
        self.num_rows_regrowth = self.num_rows_remove[name]
        grad = self.get_gradient_for_weights(weight)
        grad = grad*(new_mask==0).float()
        grad = grad.cpu().detach().numpy()
        row_l2_norm_grad = LA.norm(grad, 2, axis=1)
        row_l2_norm_grad_sorted = np.argsort(row_l2_norm_grad)[::-1][:len(row_l2_norm_grad)]
      
        for i in range(self.num_rows_regrowth):
            # print(i, row_l2_norm_grad_sorted[i])
            new_mask[row_l2_norm_grad_sorted[i], :] = 1.0

        return new_mask


    def row_balanced_gradient_growth(self, name, new_mask, weight):
        total_regrowth_per_row = math.ceil(self.num_remove[name]/new_mask.shape[0])
        grad = self.get_gradient_for_weights(weight)
       
        grad = grad*(new_mask==0).float()
        for i in range(grad.shape[0]):
        
            y, idx = torch.sort(torch.abs(grad[i,:]), descending=True)
            new_mask.data[i,:][idx[:total_regrowth_per_row]] = 1.0
        return new_mask

    def momentum_neuron_growth(self, name, new_mask, weight):
        total_regrowth = self.num_remove[name]
        grad = self.get_momentum_for_weight(weight)

        M = torch.abs(grad)
        if len(M.shape) == 2: sum_dim = [1]
        elif len(M.shape) == 4: sum_dim = [1, 2, 3]

        v = M.mean(sum_dim).data
        v /= v.sum()

        slots_per_neuron = (new_mask==0).sum(sum_dim)

        M = M*(new_mask==0).float()
        for i, fraction  in enumerate(v):
            neuron_regrowth = math.floor(fraction.item()*total_regrowth)
            available = slots_per_neuron[i].item()

            y, idx = torch.sort(M[i].flatten())
            if neuron_regrowth > available:
                neuron_regrowth = available
            threshold = y[-(neuron_regrowth)].item()
            if threshold == 0.0: continue
            if neuron_regrowth < 10: continue
            new_mask[i] = new_mask[i] | (M[i] > threshold)

        return new_mask

    '''
                UTILITY
    '''
    def get_momentum_for_weight(self, weight):
        if 'exp_avg' in self.optimizer.state[weight]:
            adam_m1 = self.optimizer.state[weight]['exp_avg']
            adam_m2 = self.optimizer.state[weight]['exp_avg_sq']
            grad = adam_m1/(torch.sqrt(adam_m2) + 1e-08)
        elif 'momentum_buffer' in self.optimizer.state[weight]:
            grad = self.optimizer.state[weight]['momentum_buffer']
        return grad

    def get_gradient_for_weights(self, weight):
        grad = weight.grad.clone()
        return grad

    def print_nonzero_counts(self):
        for module in self.modules:
            for name, tensor in module.named_parameters():
                if name not in self.masks: continue
                mask = self.masks[name]
                num_nonzeros = (mask != 0).sum().item()
                val = '{0}: {1}->{2}, density: {3:.3f}'.format(name, self.name2nonzeros[name], num_nonzeros, num_nonzeros/float(mask.numel()))
                print(val)


        for module in self.modules:
            for name, tensor in module.named_parameters():
                if name not in self.masks: continue
                print('Death rate: {0}\n'.format(self.death_rate))
                break

    def fired_masks_update(self):
        ntotal_fired_weights = 0.0
        ntotal_weights = 0.0
        layer_fired_weights = {}
        for module in self.modules:
            for name, weight in module.named_parameters():
                if name not in self.masks: continue
                self.fired_masks[name] = self.masks[name].data.byte() | self.fired_masks[name].data.byte()

                ntotal_fired_weights += float(self.fired_masks[name].sum().item())
                ntotal_weights += float(self.fired_masks[name].numel())
                layer_fired_weights[name] = float(self.fired_masks[name].sum().item())/float(self.fired_masks[name].numel())

        total_fired_weights = ntotal_fired_weights/ntotal_weights
        return layer_fired_weights, total_fired_weights